// ---------------------------- Includes -------------------------- //

// Include for console logging.
#include <map>      //We use a map for the rules
#include <vector>   //Vectors are always usefule
#include <stack>    //Stack to save and load states
#include <string>   //String parsing
#include <random>   //Random numbers
#include <sstream>  //String stream

// Use mathematics routines
#include <math.h>

// Include Allegro headers.
#include <allegro5/allegro.h>
#include <allegro5/allegro_primitives.h>
#include <allegro5/allegro_color.h>

#define PI 3.14159265

// ---------------------------- Global Allegro variables -------------------------- //
ALLEGRO_DISPLAY*     display;
ALLEGRO_EVENT_QUEUE* event_queue;
ALLEGRO_EVENT        event;
ALLEGRO_COLOR        color;

// ---------------------------- Forward declarations -------------------------- //
void init();
void deinit();
void event_loop();

std::string iterateSystem(std::string state, int iterations);
std::string rewriteSystem(std::string state);
std::string replaceRuleParameters(std::string rule);
void drawSystem(std::string state, float* position);

// ---------------------------- Global variables -------------------------- //

/**
 * This struct specifies a rules right hand side.
 * Each right hand sides have a probability and a string value.
 */
struct Rule {
    float probability;
    std::string value;

    Rule(float probability, std::string value) {
        this->probability = probability;
        this->value = value;
    }
};

//Here we will hold a map of the rules of our grammar.
std::map<char, std::vector<Rule>> rules = std::map<char, std::vector<Rule>>();

//Here we will hold a vector of generated systems.
std::vector<std::string> systems = std::vector<std::string>();

//Random number generator
std::default_random_engine generator;
std::uniform_real_distribution<float> distribution(0.0f, 1.0f);

float toRad(float degrees) {

    return degrees * PI / 180.0f;
}

/**
 * This function tries to convert a string to float and does not throw an exception if conversion fails.
 * There might be a regular stof() function in your std, but that probably throws an exception. Which
 * results in your program crashing until you create the angle reading functionality.
 */
float st2f(std::string valueStr) {
    float value;
    std::stringstream ss;
    ss << valueStr;
    ss >> value;

    return value;
}


/**
 * --Task--
 * Create the rules described in the task description.
 * Alternatively, you can use your own stochastic rules, that give you your own trees.
 */
void initRules() {
    std::vector<Rule> rulesA = std::vector<Rule>();
    rulesA.push_back(Rule(1.00f, std::string("F[+(a)F]A")));

    rules.insert(std::pair<char, std::vector<Rule>>('A', rulesA));
}

/**
 * This function creates the systems and iterates them.
 */
void initSystems() {
    //--Task-- When you are ready, use those lines to create many trees.
    /*for (int i = 0; i < 4; i++) {
        systems.push_back(iterateSystem(std::string("A"), 3 + i));
        systems.push_back(iterateSystem(std::string("A"), 3 + i));
        systems.push_back(iterateSystem(std::string("A"), 3 + i));
        systems.push_back(iterateSystem(std::string("A"), 3 + i));
        systems.push_back(iterateSystem(std::string("A"), 3 + i));
    }*/

    //Remove this line then (when everything works):
    systems.push_back(iterateSystem(std::string("A"), 2));
}

/**
 * Remove the previous systems.
 * Used, when the user presses [SPACE].
 */
void clearSystems() {
    systems.clear();
}

/**
 * Rewrites the system the number of iterations times.
 */
std::string iterateSystem(std::string state, int iterations) {
    for (int i = 0; i < iterations; i++) {
        state.assign(rewriteSystem(state));
    }
    printf("Result: %s\n", state.c_str());

    return state;
}

/**
 * --Task--
 * This function should rewrite the system, by applying the stochastic rules to the entire state.
 */
std::string rewriteSystem(std::string state) {
    std::string newState = std::string();
    float probability = 0.0f;
    float runningProbability = 0.0f;
    std::string value = std::string();
    std::vector<Rule> rule;

    for (int i = 0; i < state.length(); i++) {
        if (rules.find(char(state.at(i))) != rules.end()) { //Does a rule exist for that symbol.

            //Currently we just assign the first matching rule's value (right side).
            value = rules.at(char(state.at(i))).at(0).value;

            /**
             * You should pick one of the matching rule's right sides and assign it using the probability.
             * "distribution(generator)" gives a random number from [0, 1) range.
             * You might need to also account for floating point precision (eg, if the rule probabilities are 0.33, 0.33, 0.33 and you roll a 0.991).
             * Strings have methods assign(newString) and compare("string").
             */

            //We replace the parameters inside the value
            value = replaceRuleParameters(value);

            newState.append(value);
        } else {
            //Rule for that symbol did not exist, just copy the symbol.
            newState += state.at(i);
        }
    }

    return newState;
}

/**
 * --Task--
 * Here we want to replace the parameters (currently only "a") with some concrete values.
 * In a more complex system there would be more parameters, that have dependant values.
 */
std::string replaceRuleParameters(std::string ruleRight) {
    int n = 0;
    std::string angleStr = std::string("(a)");
    std::string replaceStr = std::string();
    char buffer[60];
    while((n = ruleRight.find(angleStr, n)) != std::string::npos) {
        //Use sprintf() function to write a random float, with 2 digits of precision after decimal point, into the buffer
        //You can get 2 digits of precision after decimal point, by the format "%.2f".
        //Assign the buffer into the replaceStr
        //Use a replace function on the ruleRight, to replace the "(a)", with a new value.

        //Then we move on.
        n += angleStr.size();
    }

    return ruleRight;
}

// ---------------------------- Main -------------------------- //
// The overall structure of our program is the familiar GUI event loop:

int main(){
    initRules();    // First we initialize the rules
    initSystems();  // Then the L-systems
    init();         // Initialize Allegro.
    event_loop();   // Run the event processing loop until a program is requested to quit
    deinit();       // Deinitialize

    return 0;
}


// ---------------------------- Allegro initialization -------------------------- //
// Initialize the Allegro system.

void init() {
    al_init();
    display = al_create_display(840, 620);
    event_queue = al_create_event_queue();
    al_register_event_source(event_queue, al_get_display_event_source(display));
    al_install_keyboard();
    al_register_event_source(event_queue, al_get_keyboard_event_source());
    al_init_primitives_addon();
}

// ---------------------------- Allegro deinitialization -------------------------- //
// Just the opposite of "init". Destroy whatever we created.

void deinit() {
    al_destroy_event_queue(event_queue);
    al_destroy_display(display);
}

// ---------------------------- Event loop -------------------------- //

void event_loop() {
    color = al_map_rgb(162, 40, 46);
    bool isRunning = true;

    while(isRunning) {
        bool have_event = al_wait_for_event_timed(event_queue, &event, 0.05);
        if (have_event) {
            if (event.type == ALLEGRO_EVENT_DISPLAY_CLOSE) isRunning = false; //Close
            if (event.type == ALLEGRO_EVENT_KEY_DOWN) {
                switch(event.keyboard.keycode) {
                    case ALLEGRO_KEY_SPACE: //When [SPACE] is pressed, create new systems.
                        clearSystems();
                        initSystems();
                    break;
                    case ALLEGRO_KEY_ENTER:
                        isRunning = false; //Close
                    break;
                }
            }
        }

        //Clear to white.
        al_clear_to_color(al_map_rgb(255, 255, 255));

        //This will draw the trees in sort of a grid.
        float position[2] = {0.0f, 110.0f};
        for (int i = 0; i < systems.size(); i++) {
            position[0] = 120.0f + 150.0f * (i%5);
            if (i > 0 && i % 5 == 0) {
                position[1] = 80.0f + 35.0f * i;
            }
            //Draw one tree:
            drawSystem(systems.at(i), position);
        }

        //Flip the buffers.
        al_flip_display();
   }
}

// ---------------------------- Drawing routines -------------------------- //
// Where all the drawing is performed.

//This struct specifies our turtle structure.
struct Turtle {
    float position[2];
    float direction[2];
    float thickness;
    float length;

    Turtle(float* position) {
        this->position[0] = position[0];
        this->position[1] = position[1];
        this->direction[0] = 0.0f;
        this->direction[1] = -1.0f;
        this->thickness = 4.0f;
        this->length = 20.0f;
    };
};

/**
 * --Task--
 * This function will use turtle graphics to draw the system.
 * You need to account for the different symbols in the system to control the turtle.
 * The Algorithmic Beauty of Plants provides a table of the symbols: http://algorithmicbotany.org/papers/abop/abop.pdf (PDF p221)
 * They are also described in the task description.
 */
void drawSystem(std::string state, float* position) {

    Turtle turtle = Turtle(position);
    float angle = 0.0;
    std::stack<Turtle> turtleStack = std::stack<Turtle>();
    float position2[2];

    for (int i = 0; i < state.length(); i++) {
        if (state.at(i) == 'F') {
            /**
             * Set the "position2" values based on the current turtle's position, length and direction.
             * Draw the line from "turtle.position" to "position2", use the al_draw_line() function.
             * Multiply the turtle's line length by 0.98 and thickness by 0.92 (or other numbers if you want).
             * Modify the position of the turtle.
             */
        }
        if (state.at(i) == '+') {
            angle = st2f(state.substr(i+2, 5)); //We assume, that the next 5 symbols constitute a float
            /**
             * Change the turtle's direction based on the read angle.
             */
        }
        if (state.at(i) == '-') {
            angle = -1.0 * st2f(state.substr(i+2, 5)); //We assume, that the next 5 symbols constitute a float
            /**
             * Change the turtle's direction based on the read angle.
             */
        }
        if (state.at(i) == '[') {
            /**
             * Use the stack to save the current turtle state.
             * Leonardo da Vinci has determined, that child (side) branches have a 0.707 width of the parent branch. Change the thickness accordingly.
             */
        }
        if (state.at(i) == ']') {
            /**
             * Load the last turtle from the stack. See methods top() and pop().
             */
        }
    }
}
